import json
import logging
from typing import Any, Callable, Optional, Tuple

from colorama import Fore, Style
from truss.build import create
from truss.local.local_config_handler import (
    LocalConfigHandler as TrussLocalConfigHandler,
)
from truss.truss_handle import TrussHandle

import baseten
from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api, settings
from baseten.common.core import MODEL_FILENAME, Semver, raises_api_error
from baseten.common.tar import create_tar_with_progress_bar
from baseten.common.types import TryPatchDraftTrussResponse
from baseten.common.util import base64_encoded_json_str

logger = logging.getLogger(__name__)


def _exists_model(
    model_name: Optional[str], model_id: Optional[str], models_provider: Callable
) -> Optional[Tuple[str, str]]:
    """Checks if an effective model exists for the purpose of deploy or one needs to be created.

    If model name is supplied and a model with that name exists then it's picked up. Otherwise
    working model id in the session is tried.

    Returns id of model if model exists, or None
    """
    models = models_provider()["models"]
    model_id_by_name = {model["name"]: model["id"] for model in models}
    model_name_by_id = {model["id"]: model["name"] for model in models}

    if model_name is not None:
        if model_name in model_id_by_name:
            return model_id_by_name[model_name], model_name
        else:
            logger.warning("Model name not found in deployed models")
            return None

    # No model_name supplied, try to work with working model id
    if model_id is not None:
        if model_id in model_name_by_id:
            return model_id, model_name_by_id[model_id]
        else:
            logger.warning("Working model id not found in deployed models")
            return None

    # No model name or working model
    return None


def build_truss(
    model: Any,
    target_directory: Optional[str] = None,
) -> TrussHandle:
    """
    Builds truss into target directory from in-memory model
    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a keras,
            sklearn, or pytorch model object)
        target_directory (str, optional): The local directory target for the truss.
            Otherwise a temporary directory will be generated
    Returns:
        TrussHandle
    """
    truss = create(model=model, target_directory=target_directory)

    truss_help_messages = f"""
    --------------------------------------------------------------------------------------------
    | {Fore.BLUE} Autogenerating Truss for your model, find more about Truss at https://truss.baseten.co/ |
    --------------------------------------------------------------------------------------------
    You can find your auto generated Truss at {Fore.BLUE} {str(truss.spec.truss_dir.resolve())}
    ---------------------------------------------------------------
    | {Fore.BLUE} Some useful commands to work with your Truss... |
    ---------------------------------------------------------------
    To load your Truss, you can either
    1. Access the Truss handle via the object returned by this function via {Fore.MAGENTA}deployed_model.truss
    2. Load the Truss handle via {Fore.MAGENTA}handle = truss.load('{str(truss.spec.truss_dir.resolve())}')
    To locally run your Truss before deploying, run {Fore.MAGENTA}handle.predict(test_input)
    To add a requirement to your Truss, run {Fore.MAGENTA}handle.add_python_requirement('numpy')
    Visit the docs to learn more about Truss!
    """
    for msg in truss_help_messages.splitlines():
        logger.info(msg.strip())
    return truss


@raises_api_error
def deploy_truss(
    b10_truss: TrussHandle,
    model_name: Optional[str] = None,
    semver_bump: str = Semver.MINOR.value,
    is_trusted: bool = False,
    publish: bool = False,
    external_model_version_id=None,
) -> BasetenDeployedModel:
    """
    Given an existing baseten truss object, deploy it onto the baseten infastructure.

    Args:
        b10_truss: A truss object representing a local baseten_truss directory
        model_name (str, optional): The name of the model to be created, if necessary.
        semver_bump (str, optional): The version bump for this deployment, one of 'MAJOR', 'MINOR', 'PATCH'.
        is_trusted (bool, optional): Whether or not to mark a truss as `trusted` on Baseten.
        publish (bool, optional): Whether or not to mark a truss as `published` on Baseten.
        external_model_version_id (str, optional): If you would rather track model IDs with your own system,
            pass the custom ID here.
    Returns:
        BasetenDeployedModel
    """
    if model_name is None:
        import coolname

        model_name = coolname.generate_slug(2)

    elif not model_name or model_name.isspace():
        raise ValueError("Model Names cannot be empty or spaces")

    # TODO(pankaj) The exists check after generating random name is incorrect.
    # If we happen to generate a random name that's already taken then we will
    # have unexpected behavior.
    model_info = _exists_model(model_name, baseten.working_model_id, api.models)

    if model_info is not None and not publish:
        resp = try_patch_draft_truss(model_name, b10_truss)
        if resp.succeeded or not resp.needs_full_deploy:
            if resp.succeeded:
                logger.info(resp.success_message or f"Model {model_name} patched successfully.")
            else:
                logger.error(
                    f"Failed to patch: `{resp.error_message}`. Model left in original state."
                )

            version_info = resp.model_version_info
            if version_info is None:
                raise RuntimeError("Expecting model version info to be set but found to be None")
            model_id, model_name = model_info
            model_version_id = version_info["id"]
            return BasetenDeployedModel(
                model_version_id=model_version_id, model_name=model_name, truss_handle=b10_truss
            )
        else:
            logger.info(f"Giving up on patching: {resp.error_message}")

    logger.info(f"Serializing {Fore.BLUE}{model_name}{Style.RESET_ALL} truss.")

    # If the truss is not scattered then gather is a no-op and returns path to
    # original truss. So, there's no additional cost to gather for trusses
    # without external packages.
    gathered_truss = TrussHandle(b10_truss.gather())
    temp_file = _archive_truss(b10_truss=gathered_truss)
    logger.info("Making contact with Baseten 👋 👽")
    s3_key = api.upload_model(temp_file, "tgz", MODEL_FILENAME)
    # String that can be passed through graphql api
    config = base64_encoded_json_str(gathered_truss._spec._config.to_dict())
    if not publish:
        model_id, model_version_id = deploy_draft_truss(
            model_name=model_name,
            s3_key=s3_key,
            config=config,
            is_trusted=is_trusted,
        )
    else:
        if not model_info:
            model_id, model_version_id = create_model_from_s3(
                model_name, s3_key, config, semver_bump, is_trusted, external_model_version_id
            )
            logger.info(f"Successfully registered model {Fore.BLUE}{model_name}{Style.RESET_ALL}.")
        else:
            model_id, model_name = model_info
            model_version_id = create_model_version_from_s3(
                model_id, s3_key, config, semver_bump, is_trusted, external_model_version_id
            )
            logger.info(
                f"Successfully created version {Fore.BLUE}{model_version_id}{Style.RESET_ALL} for {model_name}."
            )

    _generate_model_deploy_logs(model_id)
    return BasetenDeployedModel(
        model_version_id=model_version_id, model_name=model_name, truss_handle=b10_truss
    )


def deploy_draft_truss(
    model_name,
    s3_key,
    config,
    is_trusted=False,
):
    model_version_json = api.deploy_draft_truss(
        model_name=model_name,
        s3_key=s3_key,
        config=config,
        client_version=baseten.__version__,
        is_trusted=is_trusted,
    )
    can_scale_to_zero = api.get_organization_scalable_to_zero()
    scaling_message = (
        f"""
    \n📉 {Fore.MAGENTA}By default this model will scale down to zero after one hour without requests,
saving you money. When the model is invoked again it may take some time to start back up.
You can update model resources and autoscaling settings in your Baseten account."""
        if can_scale_to_zero
        else ""
    )

    logger.info(
        f"""
|---------------------------------------------------------------------------------------|
| Your model has been deployed as a draft. Draft models allow you to                    |
| iterate quickly during the deployment process. For more information,                  |
| read https://docs.baseten.co/models/deploying-models/client#stage-2-deploying-a-draft |
|                                                                                       |
| When you are ready to publish your deployed model as a new version,                   |
| pass publish=True to the baseten.deploy() command or publish the version              |
| from the Baseten UI. To skip the draft model step in future deployments,              |
| pass publish=True to the baseten.deploy() command.                                    |
|                                                                                       |
|---------------------------------------------------------------------------------------|

Deployed model info: \n{json.dumps(model_version_json, indent=4)}{scaling_message}
"""
    )

    model_id = model_version_json["id"]
    model_version_id = model_version_json["version_id"]
    # Set the newly created model model_version_json
    # be the working model for future commands
    baseten.working_model_id = model_id

    return model_id, model_version_id


def try_patch_draft_truss(
    model_name: str,
    truss_handle: TrussHandle,
) -> TryPatchDraftTrussResponse:
    dev_model_version_info = _get_draft_model_version_info(model_name)

    if dev_model_version_info is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response(
            f"No draft model found for model {model_name}."
        )

    # We need this check because a model that has not finished building will
    # not have Truss hash and signature.
    model_deployment_status = dev_model_version_info["current_model_deployment_status"]["status"]
    if model_deployment_status == "BUILDING_MODEL":
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=False,
            model_version_info=dev_model_version_info,
            error_message=f"model {model_name} is currently building, please retry after it's finished.",
        )

    truss_hash = dev_model_version_info["truss_hash"]
    if truss_hash is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response(
            f"No truss_hash found for model {model_name}."
        )

    truss_signature = dev_model_version_info["truss_signature"]
    if truss_signature is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response(
            f"No truss_signature found for model {model_name}."
        )

    TrussLocalConfigHandler.add_signature(truss_hash, truss_signature)

    patch_request = truss_handle.calc_patch(truss_hash)
    if patch_request is None:
        return TryPatchDraftTrussResponse.needs_full_deploy_response("unable to calculate patch.")

    if patch_request.prev_hash == patch_request.next_hash or len(patch_request.patch_ops) == 0:
        return TryPatchDraftTrussResponse.success_response(
            dev_model_version_info, success_message="No changes observed skipping deploy"
        )

    resp = api.patch_draft_truss(
        model_name=model_name,
        client_version=baseten.__version__,
        patch=patch_request,
    )
    if resp["succeeded"] is False:
        needs_full_deploy = True
        if "needs_full_deploy" in resp:
            needs_full_deploy = resp["needs_full_deploy"]
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=needs_full_deploy,
            model_version_info=dev_model_version_info,
            error_message=resp["error"],
        )

    return TryPatchDraftTrussResponse.success_response(
        dev_model_version_info,
        success_message=f"Successfully patched draft version of model `{model_name}`.",
    )


def _get_draft_model_version_info(model_name) -> Optional[dict]:
    model = api.get_model(model_name)
    versions = model["model_version"]["oracle"]["versions"]
    for version in versions:
        if version["is_draft"] is True:
            return version
    return None


def create_model_from_s3(
    model_name,
    s3_key,
    config,
    semver_bump,
    is_trusted=False,
    external_model_version_id=None,
):
    model_version_json = api.create_model_from_truss(
        model_name=model_name,
        s3_key=s3_key,
        config=config,
        semver_bump=semver_bump,
        client_version=baseten.__version__,
        is_trusted=is_trusted,
        external_model_version_id=external_model_version_id,
    )

    logger.info(f"Created model:\n{json.dumps(model_version_json, indent=4)}")
    model_id = model_version_json["id"]
    model_version_id = model_version_json["version_id"]
    # Set the newly created model model_version_json
    # be the working model for future commands
    baseten.working_model_id = model_id

    return model_id, model_version_id


def create_model_version_from_s3(
    model_id,
    s3_key,
    config,
    semver_bump,
    is_trusted=False,
    external_model_version_id=None,
):
    model_version_json = api.create_model_version_from_truss(
        model_id=model_id,
        s3_key=s3_key,
        config=config,
        semver_bump=semver_bump,
        client_version=baseten.__version__,
        is_trusted=is_trusted,
        external_model_version_id=external_model_version_id,
    )
    model_version_id = model_version_json["id"]
    return model_version_id


def _generate_model_deploy_logs(model_id):
    logger.info(f"{Fore.BLUE} Deploying model version.")
    model_version_web_url = f"{settings.get_server_url()}/models/{model_id}"
    logger.info("🏁 The model is being built and deployed right now 🏁")
    visit_message = (
        f"|  Visit {Fore.BLUE}{model_version_web_url}{Style.RESET_ALL} for deployment status  |"
    )
    visit_message_len = len(visit_message) - len(Fore.BLUE) - len(Style.RESET_ALL)
    logger.info("".join(["-" for _ in range(visit_message_len)]))
    logger.info(visit_message)
    logger.info("".join(["-" for _ in range(visit_message_len)]))
    _generate_organization_credit_logs()


def _generate_organization_credit_logs():
    credits = api.get_organization_credits()
    credit_granted_cents = credits["monetary_credit_granted"]
    if credit_granted_cents == 0:
        # If the organization has never been granted monetary credits, omit any credit-related logging
        return
    credit_balance_cents = credits["monetary_credit_balance"]
    needs_payment_method = credits["payment_method_status"] == "NEEDS_PAYMENT_METHOD"
    credit_balance_str = (
        "{:.2f}".format(round(credit_balance_cents / 100, 2))
        if credit_balance_cents != 0
        else str(credit_balance_cents)
    )
    no_credit_remaining = (
        credit_balance_cents <= 0 and credit_granted_cents != 0 and needs_payment_method
    )
    emoji_prefix = "💸"
    if credit_granted_cents == credit_balance_cents:
        emoji_prefix = "💰"
    elif no_credit_remaining:
        emoji_prefix = "🫙"
    suffix = (
        " Any previously-deployed model versions have been deactivated."
        if no_credit_remaining
        else ""
    )
    if needs_payment_method or credit_balance_cents > 0:
        logger.info(
            f"{emoji_prefix} Your workspace has ${credit_balance_str} of model resource credit remaining.{suffix}"
        )
    if needs_payment_method:
        billing_url = f"{Fore.BLUE}{settings.get_server_url()}/settings/billing{Style.RESET_ALL}"
        add_payment_method_str = (
            f"💳 Add a payment method to reactivate your models and deploy new versions: {billing_url}"
            if no_credit_remaining
            else f"💳 Add a payment method to keep your models running once credit runs out: {billing_url}"
        )
        logger.info(add_payment_method_str)


def _archive_truss(b10_truss: TrussHandle):
    try:
        truss_dir = b10_truss._spec.truss_dir
        temp_file = create_tar_with_progress_bar(truss_dir)
    except PermissionError:
        # Windows bug with Tempfile causes PermissionErrors
        temp_file = create_tar_with_progress_bar(truss_dir, delete=False)
    temp_file.file.seek(0)
    return temp_file


def build_and_deploy_truss(
    model: Any,
    model_name: Optional[str] = None,
    version_bump=Semver.MINOR.value,
    is_trusted=False,
    publish=False,
    external_model_version_id=None,
) -> BasetenDeployedModel:
    if isinstance(model, TrussHandle):
        return deploy_truss(
            model, model_name, version_bump, is_trusted, publish, external_model_version_id
        )
    truss = build_truss(model)
    return deploy_truss(
        truss, model_name, version_bump, is_trusted, publish, external_model_version_id
    )


def pull_model(model_version_id: str, directory: str = "."):
    deployed_model = BasetenDeployedModel(model_version_id=model_version_id)
    return deployed_model.pull(directory)



import functools
import json
import logging
from contextlib import contextmanager
from http import HTTPStatus
from typing import Optional

from requests.exceptions import ConnectionError, ConnectTimeout, HTTPError, Timeout

logger = logging.getLogger(__name__)


@contextmanager
def error_handling_context(error_prefix: str):
    def log_error(msg: str):
        logger.error(f"{error_prefix} {msg}")

    try:
        yield
    except HTTPError as err:
        if err.response.status_code == HTTPStatus.SERVICE_UNAVAILABLE:
            json_msg = _get_json(err.response.text)
            if json_msg is not None and isinstance(json_msg, dict) and "error" in json_msg:
                log_error(json_msg["error"])
            else:
                log_error(err.response.text)
        else:
            raise err
    except (ConnectionError, ConnectTimeout):
        log_error(
            "Unable to communicate with the backend server."
            " Please verify that you have internet connectivity."
        )
    except Timeout:
        log_error("Request timed out")


def error_handling(error_prefix: str):
    def inner(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with error_handling_context(error_prefix):
                return func(*args, **kwargs)

        return wrapper

    return inner


def _get_json(msg: str) -> Optional[dict]:
    try:
        return json.loads(msg)
    except ValueError:
        return None



"""Baseten

    isort:skip_file
"""

from truss.build import load
from pathlib import Path
from single_source import get_version
import logging
from typing import List, DefaultDict


from collections import defaultdict

from baseten.baseten_deployed_model import BasetenDeployedModel, models_summary  # noqa: E402
from baseten.common.model_deployer import (
    build_truss,
    deploy_truss,
    build_and_deploy_truss,
    pull_model,
)
from baseten.common.files import upload_dataset
from baseten.common.util import setup_logger  # noqa: E402
from baseten.common import settings  # noqa: E402


__version__ = get_version(__name__, Path(__file__).parent.parent)


def version():
    return __version__


patched_modules: DefaultDict = defaultdict(lambda: [])


logger = logging.getLogger(__name__)
setup_logger("baseten", logging.INFO)
if settings.DEBUG:
    setup_logger("baseten", logging.DEBUG)
logger.debug(f"Starting the client with the server URL set to {settings.get_server_url()}")

# The Baseten model ID that either a) the user initialized baseten with, or b) was created when deploying a model.
working_model_id = None

deploy = build_and_deploy_truss  # This allows the user to call baseten.deploy(model)
build_truss = build_truss
deploy_truss = deploy_truss
upload = upload_dataset
load_truss = load
pull = pull_model


models_summary = models_summary


def set_log_level(level):
    setup_logger("baseten", level)


def init(baseten_model_id=None):
    """Initialize Baseten

    Args:
        baseten_model_id (str, optional): The Baseten model id to initialize the client with.
        If not provided, a new Baseten model will be created at deploy() time.
    """
    global working_model_id
    working_model_id = baseten_model_id


def deployed_model_id(model_id: str) -> BasetenDeployedModel:
    """Returns a BasetenDeployedModel object for interacting with the model model_id.

    Args:
        model_id (str): The ID of the model

    Returns:
        BasetenDeployedModel (BasetenDeployedModel): An object for interacting with the model
    """
    return BasetenDeployedModel(model_id=model_id)


def deployed_model_version_id(model_version_id: str) -> BasetenDeployedModel:
    """Returns a BasetenDeployedModel object for interacting with the model version model_version_id.

    Args:
        model_version_id (str): The ID of a model version

    Returns:
        BasetenDeployedModel (BasetenDeployedModel): An object for interacting with the model
    """
    return BasetenDeployedModel(model_version_id=model_version_id)


def deployed_external_model_version_id(external_model_version_id: str) -> BasetenDeployedModel:
    """Returns a BasetenDeployedModel object for interacting with the model version external_model_version_id.

    Args:
        external_model_version_id (str): The non-Baseten-generated ID assigned during model deployment

    Returns:
        BasetenDeployedModel (BasetenDeployedModel): An object for interacting with the model
    """
    return BasetenDeployedModel(external_model_version_id=external_model_version_id)


def login(api_key: str):
    """Set the API key for the client.

    Args:
        api_key (str): Your API key
    """
    settings.set_config_value("api", "api_key", api_key)
    logger.info("API key set.")


def set_jwt(jwt: str):
    """Sets a JWT for the client. For internal use only.

    Args:
        jwt (str)
    """

    settings.set_jwt_auth_token(jwt)
    logger.info("JWT set.")


def configure(server_url: str):
    """Configure client to use on-prem hosted Baseten environment.

    !!! note

        The client works out of the box without any need to run this command.
        This command is only required if you have Baseten Enterprise hosted on-prem and is not
        required for most users.

    Args:
        server_url (str): The base URL of the server
    """
    if settings.set_server_url(server_url):
        logger.info("Saved server URL.")
    else:
        logger.info("That is not a valid URL.")


def route(
    path: str,
    is_public: bool = False,
    allowed_domains: List[str] = [],
    allowed_methods: List[str] = ["GET"],
):
    """
    Decorator: transform a Python function in Blueprint into an endpoint.

    **Example:**

    ```python
    @route(path="/some-route")
    def my_route(request):
        ...
    ```

    ```python
    @route(
        path="/public-route",
        is_public=True,
        allowed_domains=["https://example.com", "https://another-example.com"],
        allowed_methods=["GET", "POST"]
    )
    def my_route(request):
        ...
    ```

    Args:
        path: Path to use for the route
        is_public: is this a public endpoint?
            (Public endpoints do not require an API key to request.)
        allowed_domains: Allowed domains prevent CORS issues when invoking public endpoints.
        allowed_methods: Choose to support `'GET'`, `'POST'`, `'PUT'`, and `'DELETE'` methods.
    """

    def route_decorator(func):
        return func

    return route_decorator



import functools
import json
from enum import Enum

from requests import RequestException


class Error(Exception):
    """Base Baseten Error"""

    def __init__(self, message):
        super().__init__(message)
        self.message = message


class AuthorizationError(Error):
    """Raised in places where the user needs to be logged in and is not."""

    pass


class FrameworkNotSupportedError(Error):
    """Raised in places where the user attempts to use Baseten with an unsupported framework"""

    pass


class ModelNotSupportedError(Error):
    """Raised when a user attempts to use unsupported models for a
    framework that is otherwise supported"""

    pass


class ModelFilesMissingError(Error):
    """Raised when a user attempts to deploy a model which requires supporting files."""

    pass


class ModelClassImplementationError(Error):
    """Raised when a user attempts to deploy a model class which does not implement necessary methods."""

    pass


class SampleDataInputShapeError(Error):
    """Raised when deploying a model whose input shape doesn't match the number of feature names given by the user."""

    pass


class TrussBuildError(Error):
    """Raised when a user attempts to build a Truss and it failts"""

    pass


class TrussDeployError(Error):
    """Raised when a user attempts to deploy a Truss to Baseten and it fails"""

    pass


class InvalidDatasetError(Error):
    """Raised when user attempts to upload an invalid dataset for training"""

    pass


class ApiError(Error):
    """Errors in calling the Baseten API."""

    def __init__(self, message, response=None):
        super().__init__(message)
        self.response = response

    def __str__(self):
        error_str = self.message
        if self.response is not None:  # non-200 Response objects are falsy, hence the not None.
            error_message = json.loads(self.response.content)
            error_message = error_message["error"] if "error" in error_message else error_message
            error_str = f"{error_str}\n<Server response: {error_message}>"
        return error_str


def raises_api_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RequestException as err:
            raise ApiError(f"Error calling {func.__name__}.", response=err.response) from err

    return wrapper


class Semver(Enum):
    MAJOR = "MAJOR"
    MINOR = "MINOR"
    PATCH = "PATCH"


MODEL_FILENAME = "model"



import os

import click

import baseten


@click.command()
@click.argument("target_directory", required=False)
@click.argument("model_name", required=False)
def deploy_from_directory(target_directory=None, model_name=None):
    """Deploys a Truss from a directory

    TARGET_DIRECTORY (Str): Directory of the truss

    MODEL_NAME (Str): Name of the model
    """
    if target_directory is None:
        target_directory = os.getcwd()
    truss = baseten.load_truss(target_directory)
    baseten.deploy_truss(truss, model_name=model_name)



import logging
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Dict, List, Optional

from baseten.common import settings
from baseten.common.api import (
    cancel_training_run,
    deploy_from_training_run,
    finetune_zoo_model,
    get_all_training_runs,
    get_blueprint_projects,
    get_training_run_by_id,
    get_training_run_by_name,
)
from baseten.common.files import DatasetTrainingType
from baseten.models.flan_t5 import FlanT5
from baseten.models.foundational_model import FoundationalModel
from baseten.models.llama import Llama
from baseten.models.stable_diffusion import StableDiffusionPipeline
from baseten.training.datasets import DatasetIdentifier
from baseten.training.logs import TrainingLogsConsumer
from baseten.training.utils import encode_base64_json

logger = logging.getLogger(__name__)


class FinetuningConfig(ABC):
    """
    Base abstract class for defining fine tuning configs. Each new fine-tuning
    algorithm that we support must support a subclass of this.

    Each fine-tuning algorithm will have a different set of variables that
    need to be provided.
    """

    @property
    @abstractmethod
    def dataset_training_type(self) -> DatasetTrainingType:
        """
        The dataset training type associated with this config type.
        """

    @property
    @abstractmethod
    def training_truss_name(self) -> str:
        """
        The name of the truss associated with this config.
        """

    @abstractmethod
    def resolve_fine_tuning_variables(self) -> Dict:
        """
        Method that must be implemented that returns a variables dictionary
        for this type of FineTuning configuration.
        """


@dataclass
class LlamaConfig(FinetuningConfig):
    """
    Training config for LLaMA-7B.

    **Examples:**

    ```python
    from baseten.training import Dataset, LlamaConfig

    config = LlamaConfig(
        input_dataset=Dataset("DATASET_ID"),
        epochs=3,
        learning_rate=5e-5,
        max_steps=1000,
        train_batch_size=8,
        sample_batch_size=8,
        report_to="wandb"
    )
    ```

    Args:
        model_id:
            Pretrained model to fine-tune
        source_col_name:
            Name of the source column in the input CSV
        target_col_name:
            Name of the target column in the input CSV
        evaluation_strategy:
            Interval for evaluation (default: "epoch")
        train_batch_size:
            Batch size for training (default: 8)
        train_micro_batch_size:
            Micro batch size for training (default: 4)
        sample_batch_size:
            Batch size for sampling (default: 8)
        sample_micro_batch_size:
            Micro batch size for sampling (default: 4)
        gradient_accumulation:
            Whether to perform gradient accumulation (default: True)
        gradient_checkpointing:
            Whether to use gradient checkpointing (default: False)
        learning_rate:
            Learning rate for optimizer (default: 5e-5)
        weight_decay:
            Weight decay for optimizer (default: 0.0)
        adam_beta1:
            Beta1 parameter for AdamW optimizer (default: 0.9)
        adam_beta2:
            Beta2 parameter for AdamW optimizer (default: 0.999)
        adam_epsilon:
            Epsilon parameter for AdamW optimizer (default: 1e-8)
        max_grad_norm:
            Maximum gradient norm (default: 1.0)
        epochs:
            Number of epochs to train for (default: 3.0)
        max_steps:
            Maximum number of steps to train for (default: -1)
        warmup_steps:
            Number of warmup steps (default: 0)
        logging_steps:
            Interval for logging (default: 500)
        seed:
            Random seed (default: 42)
        fp16:
            Whether to use mixed precision training (default: True)
        run_name:
            Name for the run (default: "llama-7b")
        disable_tqdm:
            Whether to disable tqdm progress bars (default: True)
        label_smoothing_factor:
            The label smoothing factor to use (default: 0.0)
        adafactor:
            Whether to use the Adafactor optimizer (default: False)
        report_to:
            Destination for metrics reporting (default: None)
        max_length:
            Maximum input length (default: 256)
        lora_r:
            Number of attention heads for LORA (default: 8)
        lora_target_modules:
            List of target modules for LORA (default: ["q_proj", "v_proj"])
        lora_alpha:
            Alpha parameter for LORA (default: 16.0)
        lora_dropout:
            Dropout probability for LORA (default: 0.05)
        lora_bias:
            Bias initialization for LORA (default: "none")
    """

    input_dataset: DatasetIdentifier
    model_id: str = "decapoda-research/llama-7b-hf"
    evaluation_strategy: str = "epoch"
    train_batch_size: int = 16
    train_micro_batch_size: int = 8
    sample_batch_size: int = 16
    sample_micro_batch_size: int = 8
    gradient_accumulation: bool = True
    gradient_checkpointing: bool = False
    learning_rate: float = 5e-5
    weight_decay: float = 0.0
    adam_beta1: float = 0.9
    adam_beta2: float = 0.999
    adam_epsilon: float = 1e-8
    max_grad_norm: float = 1.0
    epochs: float = 3.0
    max_steps: int = -1
    warmup_steps: int = 0
    logging_steps: int = 10
    eval_steps: int = 100
    seed: int = 42
    fp16: bool = True
    run_name: str = "blueprint_llama_7b"
    report_to: str = "wandb"
    wandb_api_key: Optional[str] = None
    disable_tqdm: bool = False
    label_smoothing_factor: float = 0.0
    adafactor: bool = False
    source_col_name: str = "source"
    target_col_name: str = "target"
    max_length: int = 512
    mask_source: bool = True
    lora_r: int = 8
    lora_target_modules: list = field(default_factory=lambda: ["q_proj", "v_proj"])
    lora_alpha: float = 16.0
    lora_dropout: float = 0.05
    lora_bias: str = "none"

    def _resolve_dataset(self) -> str:
        return self.input_dataset.resolve_dataset_url(training_type=self.dataset_training_type)

    def resolve_fine_tuning_variables(self) -> Dict:
        """
        Returns fine-tuning variables dict to send when triggering the fine-tuning
        job. The dataset URLs here are resolved to S3 URLs.
        """
        resolved_variables = asdict(self)

        resolved_variables.update({"dataset_zip_url": self._resolve_dataset()})

        del resolved_variables["input_dataset"]

        return {key: value for key, value in resolved_variables.items() if value is not None}

    @property
    def training_truss_name(self) -> str:
        return "llama7b"

    @property
    def dataset_training_type(self) -> DatasetTrainingType:
        return DatasetTrainingType.LLAMA_7B


@dataclass
class FlanT5BaseConfig(FinetuningConfig):
    """
    Fine-tuning config for Flan-T5 fine-tuning on the Base version of the model.

    **Examples:**

    ```python
    from baseten.training import Dataset, FlanT5BaseConfig

    config = FlanT5BaseConfig(
        input_dataset=Dataset("DATASET_ID"),
        epochs=1,
        learning_rate=0.00003
    )
    ```

    Args:
        input_dataset:
            An identifier, either an ID or a public URL, for the Dataset that Flan-T5 should use
        wandb_api_key:
            [API key](https://docs.wandb.ai/quickstart) for Weights & Biases
            to monitor your model training
        model_id:
            Base model to train (default: "google/flan-t5-base")
        epochs:
            Number of epochs to run
        train_batch_size:
            Batch size to use for training
        sample_batch_size:
            Batch size to use for sampling
        generation_max_length:
            Maximum length of the generated text during evaluation
        generation_num_beams:
            Number of beams to use for generation
        learning_rate:
            The learning rate to use for training
        gradient_checkpointing:
            Whether to use gradient checkpointing, which can reduce memory usage
        logging_steps:
            Number of steps between logs
        source_col_name:
            Name of the source column in the input CSV
        target_col_name:
            Name of the target column in the input CSV
        weight_decay:
            Weight decay to use for training
        adam_beta1:
            Beta1 parameter for AdamW optimizer
        adam_beta2:
            Beta2 parameter for AdamW optimizer
        adam_epsilon:
            Epsilon parameter for AdamW optimizer
        max_grad_norm:
            Maximum gradient norm to use for training
        lr_scheduler_type:
            Type of learning rate scheduler to use. Options:

            - "linear"
            - "cosine"
            - "cosine_with_restarts"
            - "polynomial"
            - "constant"
            - "constant_with_warmup"
            - "inverse_sqrt"
        warmup_steps:
            Number of steps used for a linear warmup from 0 to learning_rate. Overrides any effect of warmup_ratio.
        warmup_ratio:
            Ratio of total training steps used for a linear warmup from 0 to learning_rate.
        optimizer:
            Optimizer to use for training. Options:

            - "adamw_hf"
            - "adamw_torch"
            - "adamw_apex_fused"
            - "adamw_anyprecision"
            - "adafactor"
        label_smoothing_factor:
            The label smoothing factor to use. Zero means no label smoothing,
            otherwise the underlying oneshot-encoded labels are changed from 0s
            and 1s to label_smoothing_factor/num_labels and 1 -
            label_smoothing_factor + label_smoothing_factor/num_labels
            respectively.
        metric:
            Metric to use for evaluation (options include all metrics supported by the Huggingface Evaluate library)
    """

    input_dataset: DatasetIdentifier
    wandb_api_key: Optional[str] = None
    model_id: str = "google/flan-t5-base"
    epochs: int = 1
    train_batch_size: int = 16
    sample_batch_size: int = 16
    generation_max_length: int = 140
    generation_num_beams: int = 1
    learning_rate: float = 3e-5
    gradient_checkpointing: bool = True
    logging_steps: int = 10
    source_col_name: str = "source"
    target_col_name: str = "target"
    weight_decay: float = 0.0
    adam_beta1: float = 0.9
    adam_beta2: float = 0.999
    adam_epsilon: float = 1e-8
    max_grad_norm: float = 1.0
    lr_scheduler_type: str = "linear"
    warmup_steps: int = 0
    warmup_ratio: float = 0.0
    optimizer: str = "adamw_hf"
    label_smoothing_factor: float = 0.0
    metric: str = "rouge"

    def _resolve_dataset(self) -> str:
        return self.input_dataset.resolve_dataset_url(training_type=self.dataset_training_type)

    def resolve_fine_tuning_variables(self) -> Dict:
        """
        Returns fine-tuning variables dict to send when triggering the fine-tuning
        job. The dataset URLs here are resolved to S3 URLs.
        """
        resolved_variables = asdict(self)

        resolved_variables.update({"dataset_zip_url": self._resolve_dataset()})

        del resolved_variables["input_dataset"]

        return {key: value for key, value in resolved_variables.items() if value is not None}

    @property
    def training_truss_name(self) -> str:
        return "flan_t5_base"

    @property
    def dataset_training_type(self) -> DatasetTrainingType:
        return DatasetTrainingType.FLAN_T5


@dataclass
class DreamboothConfig(FinetuningConfig):
    """
    Fine-tuning config for Dreambooth fine-tuning with Stable Diffusion.

    **Examples:**

    ```python
    from baseten.training import Dataset, DreamboothConfig

    config = DreamboothConfig(
        instance_prompt="photo of olliedog", # Dog's name is "Ollie"
        input_dataset=Dataset("DATASET_ID"),
        class_prompt="a photo of a dog",
        num_train_epochs=10
    )
    ```

    Args:
        instance_prompt:
            The prompt with an identifier specifying the instance
            that you're teaching Stable Diffusion
        input_dataset:
            An identifier, either an ID or a public URL, for the Dataset that Dreambooth should use
        wandb_api_key:
            [API key](https://docs.wandb.ai/quickstart) for Weights & Biases
            to monitor your model training
        pretrained_model_name_or_path:
            Path to pretrained model or model identifier from huggingface.co/models.
        revision:
            Revision of pretrained model identifier from huggingface.co/models.
        tokenizer_name:
            Pretrained tokenizer name or path if not the same as model_name
        class_prompt:
            The prompt to specify images in the same class as
            your instance images. This helps regularize the model
            (e.g. so that not all prompts with "dog" look
            like your dog but only "olliedog" does)
        with_prior_preservation:
            Flag to use prior preservation loss
        prior_loss_weight:
            The weight of the prior preservation loss
        gradient_accumulation_steps:
            Number of gradient accumulation steps to use for training. Defaults to 1.
        num_class_images:
            The number of class images to use for fine-tuning,
            only relevant if using prior preservation. If greater
            than the number of class images in your dataset, it
            will generate the remainder using the base model.
        seed:
            The random seed to use for fine-tuning
        resolution:
            The resolution of your input images. Images will be resized to this value.
        center_crop:
            Whether to center crop the images to the resolution
        train_text_encoder:
            Whether to train the text encoder alongside the UNet
        train_batch_size:
            The batch size to use for training. This value can cause OOMs if
            too large. We recommend using between 1-4 based on the
            resolution of your images.
        sample_batch_size:
            The batch size to use for sampling images.
        num_train_epochs:
            The number of epochs to train for. If you set max_train_steps, this will be ignored.
        max_train_steps:
            The number of training steps to train for.
            If you set this, num_train_epochs will be ignored.
            Defaults to 1000 train steps.
        learning_rate:
            Initial learning rate (after the potential warmup period) to use.
        lr_scheduler:
            The scheduler type to use. Choose between ["linear",
            "cosine", "cosine_with_restarts", "polynomial",
            "constant", "constant_with_warmup"].
        lr_warmup_steps:
            The number of steps to warmup for the learning rate schedule.
        adam_beta1:
            The beta1 parameter for the Adam optimizer.
        adam_beta2:
            The beta2 parameter for the Adam optimizer.
        adam_weight_decay:
            The weight decay value for the Adam optimizer.
        adam_epsilon:
            The epsilon value for the Adam optimizer.
        max_grad_norm:
            The max gradient normalization to clip gradients to.
            Helps prevent exploding gradients.
        mixed_precision:
            Whether to use mixed precision. Choose between
            fp16 and bf16 (bfloat16) or None for no mixed
            precision. This value can cause OOMs in some
            cases with batch size. We recommend using fp16.
        image_log_steps:
            The number of steps to log sample images to Weights
            and Biases. This allows you to visually assess your
            model during training. Only relevant if wandb_api_key is set.
    """

    instance_prompt: str
    input_dataset: DatasetIdentifier
    wandb_api_key: Optional[str] = None
    hf_access_token: Optional[str] = None

    pretrained_model_name_or_path: str = "CompVis/stable-diffusion-v1-4"
    revision: Optional[str] = None
    tokenizer_name: Optional[str] = None

    class_prompt: Optional[str] = None
    with_prior_preservation: bool = False
    gradient_checkpointing: bool = True
    gradient_accumulation_steps: int = 1
    prior_loss_weight: float = 1.0
    num_class_images: int = 100
    seed: Optional[int] = None
    resolution: int = 512
    center_crop: bool = False
    train_text_encoder: bool = False
    train_batch_size: int = 1
    sample_batch_size: int = 1
    num_train_epochs: int = 1
    max_train_steps: int = 1000
    learning_rate: float = 1e-6
    lr_scheduler: str = "constant"
    lr_warmup_steps: int = 500
    adam_beta1: float = 0.9
    adam_beta2: float = 0.999
    adam_weight_decay: float = 1e-2
    adam_epsilon: float = 1e-08
    max_grad_norm: float = 1.0
    mixed_precision: str = "fp16"
    image_log_steps: int = 20

    def _resolve_dataset(self) -> str:
        return self.input_dataset.resolve_dataset_url(training_type=self.dataset_training_type)

    def resolve_fine_tuning_variables(self) -> Dict:
        """
        Returns fine-tuning variables dict to send when triggering the fine-tuning
        job. The dataset URLs here are resolved to S3 URLs.
        """
        resolved_variables = asdict(self)

        resolved_variables.update({"dataset_zip_url": self._resolve_dataset()})

        del resolved_variables["input_dataset"]

        return {key: value for key, value in resolved_variables.items() if value is not None}

    @property
    def training_truss_name(self) -> str:
        return "dreambooth"

    @property
    def dataset_training_type(self) -> DatasetTrainingType:
        return DatasetTrainingType.DREAMBOOTH


@dataclass
class FullStableDiffusionConfig(FinetuningConfig):
    """
    Fine-tuning config for full fine-tuning with Stable Diffusion.

    **Examples:**

    ```python
    from baseten.training import Dataset, FullStableDiffusionConfig

    config = FullStableDiffusionConfig(
        input_dataset=Dataset("DATASET_ID"),
        train_text_encoder=False,
        num_train_epochs=10
    )
    ```

    Args:
        input_dataset:
            An identifier, either an ID or a public URL, for the Dataset that Dreambooth should use
        wandb_api_key:
            [API key](https://docs.wandb.ai/quickstart) for Weights & Biases to monitor your model training
        pretrained_model_name_or_path:
            Path to pretrained model or model identifier from huggingface.co/models.
        revision:
            Revision of pretrained model identifier from huggingface.co/models.
        seed:
            The random seed to use for fine-tuning
        gradient_accumulation_steps:
            Number of gradient accumulation steps to use. Defaults to 1.
        resolution:
            The resolution of your input images. Images will be resized to this value.
        num_train_epochs:
            The number of epochs to train for. If you set max_train_steps, this will be ignored.
        train_text_encoder:
            Whether to train the text encoder alongside the UNet
        center_crop:
            Whether to center crop the images to the resolution
        random_flip:
            Whether to randomly flip the images horizontally
        train_batch_size:
            The batch size to use for training. This value can cause OOMs if too large.
            We recommend using between 1-4 based on the resolution of your images.
        max_train_steps:
            The number of training steps to train for. If you set this, num_train_epochs will be ignored.
        learning_rate:
            Initial learning rate (after the potential warmup period) to use.
        lr_scheduler:
            The scheduler type to use. Choose between ["linear", "cosine",
            "cosine_with_restarts", "polynomial", "constant", "constant_with_warmup"].
        lr_warmup_steps:
            The number of steps to warmup for the learning rate schedule.
        use_ema:
            Whether to use exponential moving average for the model weights.
        non_ema_revision:
            Revision of pretrained non-ema model identifier. Must be remote
            repository specified with --pretrained_model_name_or_path.
        adam_beta1:
            The beta1 parameter for the Adam optimizer.
        adam_beta2:
            The beta2 parameter for the Adam optimizer.
        adam_weight_decay:
            The weight decay value for the Adam optimizer.
        adam_epsilon:
            The epsilon value for the Adam optimizer.
        max_grad_norm:
            The max gradient normalization to clip gradients to. Prevents exploding gradients.
        image_log_steps:
            The number of steps to log sample images to Weights and Biases. This allows you to visually assess your
            model during training. Only relevant if wandb_api_key is set.
    """

    input_dataset: DatasetIdentifier
    wandb_api_key: Optional[str] = None
    hf_access_token: Optional[str] = None

    pretrained_model_name_or_path: str = "CompVis/stable-diffusion-v1-4"
    revision: Optional[str] = None
    max_train_samples: Optional[int] = None
    seed: Optional[int] = None
    resolution: int = 512
    num_train_epochs: int = 10
    train_text_encoder: bool = False
    center_crop: bool = False
    random_flip: bool = False
    train_batch_size: int = 1
    max_train_steps: Optional[int] = None
    gradient_checkpointing: bool = True
    gradient_accumulation_steps: int = 1
    learning_rate: float = 1e-5
    lr_scheduler: str = "cosine"
    lr_warmup_steps: int = 500
    use_ema: bool = False
    non_ema_revision: Optional[str] = None
    adam_beta1: float = 0.9
    adam_beta2: float = 0.999
    adam_weight_decay: float = 1e-2
    adam_epsilon: float = 1e-08
    max_grad_norm: float = 1.0
    image_log_steps: int = 50

    def _resolve_dataset(self) -> str:
        return self.input_dataset.resolve_dataset_url(training_type=self.dataset_training_type)

    def resolve_fine_tuning_variables(self) -> Dict:
        """
        Returns fine-tuning variables dict to send when triggering the fine-tuning
        job. The dataset URLs here are resolved to S3 URLs.
        """
        resolved_variables = asdict(self)

        resolved_variables.update({"dataset_zip_url": self._resolve_dataset()})

        del resolved_variables["input_dataset"]

        return {key: value for key, value in resolved_variables.items() if value is not None}

    @property
    def training_truss_name(self) -> str:
        return "full_stable_diffusion"

    @property
    def dataset_training_type(self) -> DatasetTrainingType:
        return DatasetTrainingType.CLASSIC_STABLE_DIFFUSION


class FinetuningRun:
    """
    Class to represent a single finetuning run.

    **Examples:**

    Instantiate a new fine-tuning run and stream logs

    ```python
    from baseten.training import FinetuningRun

    my_run = FinetuningRun.create(
        trained_model_name="My Model",
        fine_tuning_config=config
    )
    my_run.stream_logs()
    ```

    Access an existing fine-tuning run

    ```python
    from baseten.training import FinetuningRun

    my_run = FinetuningRun("RUN_ID")
    ```
    """

    @staticmethod
    def create(
        trained_model_name: str,
        fine_tuning_config: FinetuningConfig,
        auto_deploy: bool = True,
        verbose: bool = True,
    ) -> Optional["FinetuningRun"]:
        """
        Fine-tune a model by creating a FinetuningRun

        **Example:**

        ```python
        from baseten.training import FinetuningRun

        my_run = FinetuningRun.create(
            trained_model_name="My Model",
            fine_tuning_config=config
        )
        ```

        Args:
            trained_model_name: The name you want your fine-tuned model to have
            fine_tuning_config: The configuration for the fine-tuning process
            auto_deploy: Flag for whether or not the model should be deployed
                after finetuning is complete.
            verbose: Toggle this for verbose output

        Returns:
            FinetuningRun: The newly created FinetuningRun on Blueprint
        """
        variables_dict = fine_tuning_config.resolve_fine_tuning_variables()
        encoded_variables = encode_base64_json(variables_dict)

        logger.info("Starting fine-tuning of %s", fine_tuning_config.training_truss_name)
        run_id = finetune_zoo_model(
            trained_model_name,
            fine_tuning_config.training_truss_name,
            encoded_variables,
            auto_deploy,
        )

        run = FinetuningRun(run_id)
        if verbose:
            logger.info("🔮 Next steps:")
            logger.info("\t* 👽 Visit %s to see the progress of your run" % run.blueprint_url)
            logger.info("\t* 🪵  Run the following Python code to stream the logs here")
            logger.info('\t\t* run = FinetuningRun("%s")' % run.id)
            logger.info("\t\t* run.stream_logs()")
            logger.info(
                "\t* 📫 You'll get an email once the finetuning run completes, "
                "and another when your model is deployed."
            )
        else:
            logger.info(
                "\t* 📫 You'll get an email once the finetuning run completes, and will then"
                "be able to deploy it."
            )
        return run

    @staticmethod
    def list() -> List["FinetuningRun"]:
        """
        List all finetuning runs for user.

        **Example**

        ```python
        from baseten.training import FinetuningRun

        FinetuningRun.list()
        ```

        """
        resolved_runs = get_all_training_runs()
        return list([FinetuningRun(r["id"], r["trained_model_name"]) for r in resolved_runs])

    class Status(Enum):
        """Status of FinetuningRun on Baseten."""

        PENDING = "PENDING"
        RUNNING = "RUNNING"
        SUCCEEDED = "SUCCEEDED"
        FAILED = "FAILED"
        CANCELLED = "CANCELLED"

    def __init__(
        self,
        id: Optional[str] = None,
        trained_model_name: Optional[str] = None,
    ):
        """Initialize the object with the given id.

        Args:
            id: The ID of the FinetuningRun on Blueprint
            trained_model_name: Name of the FinetuningRun on Blueprint
        """
        if id is None and trained_model_name is None:
            raise ValueError("Most provide either `id` or `trained_model_name`")
        if trained_model_name is not None:
            self.refresh(trained_model_name=trained_model_name)
        else:
            self.refresh(id=id)

    def __repr__(self) -> str:
        trained_name = ""
        if self.trained_model_name:
            trained_name = f", name: {self.trained_model_name}"
        return f"""<FinetuningRun id: {self.id}{trained_name}>"""

    @property
    def status(self) -> str:
        """
        Get the status of your FinetuningRun.
        Statuses are:

        1. PENDING: The run has not yet started
        2. RUNNING: The run is actively fine-tuning the model
        3. SUCCEEDED: The run is finished and fine-tuned the model
        4. FAILED: The run hit an error and did not fine-tune the model
        5. CANCELLED: The run was cancelled by a user

        **Example**

        ```python
        my_run = FinetuningRun("RUN_ID")
        my_run.status
        ```
        """
        return self._status

    @property
    def is_pending(self) -> bool:
        """Is the run pending (waiting to start)"""
        return self._status == FinetuningRun.Status.PENDING.value

    @property
    def is_running(self) -> bool:
        """Is the run running (actively fine-tuning the model)"""
        return self._status == FinetuningRun.Status.RUNNING.value

    @property
    def is_succeeded(self) -> bool:
        """Has the run succeeded (finished fine-tuning the model)"""
        return self._status == FinetuningRun.Status.SUCCEEDED.value

    @property
    def is_failed(self) -> bool:
        """Has the run failed (errored and stopped running)"""
        return self._status == FinetuningRun.Status.FAILED.value

    @property
    def is_cancelled(self) -> bool:
        """Was the run cancelled (terminated by the user)"""
        return self._status == FinetuningRun.Status.CANCELLED.value

    def refresh(self, id: Optional[str] = None, trained_model_name: Optional[str] = None):
        """Re-fetch training data from the server."""
        if id is not None:
            training_run_data = get_training_run_by_id(id)
        elif trained_model_name is not None:
            training_run_data = get_training_run_by_name(trained_model_name)
        else:
            training_run_data = get_training_run_by_id(self.id)

        # TODO (Sid): Handle case of training run not found

        self.truss_name = training_run_data["truss_name"]
        self._status = training_run_data["status"]
        self.created = training_run_data["created"]
        self.started = training_run_data["started"]
        self.stopped = training_run_data["stopped"]
        self.trained_model_name = training_run_data["trained_model_name"]
        self.id: str = training_run_data["id"]

        if training_run_data["latest_oracle_version"]:
            self.deployed_model_id = training_run_data["latest_oracle_version"]["oracle"]["id"]
        else:
            self.deployed_model_id = None

        self.variables = {
            variable["key"]: variable["value"] for variable in training_run_data["variables"]
        }

    # Mutations
    def deploy(self, idle_time_minutes: int = 30, verbose: bool = True) -> FoundationalModel:
        """
        Deploy the fine-tuned model created during the FinetuningRun

        **Example**

        ```python
        from baseten.training import FinetuningRun

        my_run = FinetuningRun("RUN_ID")
        # After the run is finished
        my_run.deploy()
        ```

        Args:
            idle_time_minutes: How long the deployed model should wait between invocations
                before scaling resources to zero
            verbose: Toggle this for verbose output

        Returns:
            StableDiffusionPipeline: A model object using the finetuning run results
        """
        model_id, _ = deploy_from_training_run(
            training_run_id=self.id,
            name=self.trained_model_name,
            idle_time_minutes=idle_time_minutes,
        )

        if verbose:
            logger.info("🔮 Deploying finetuning run %s..." % self.id)
            logger.info(
                "\t* 👽 Visit %s to see the progress of your deployment"
                % self.blueprint_deployed_model_url
            )
            logger.info(
                "\t* 📫 You'll get an email once the deploy completes, and will then be able to invoke your model."
            )

        # TODO: determine which model to return based on finetuning type
        return _fetch_model_pipeline(self.truss_name, model_id)

    def cancel(self) -> bool:
        """
        Cancels this FinetuningRun.

        **Example**

        ```python
        from baseten.training import FinetuningRun

        my_run = FinetuningRun("RUN_ID")
        # While the run is still going
        my_run.cancel()
        ```

        Returns:
            bool: True if the run was successfully canceled
        """
        return cancel_training_run(training_run_id=self.id)

    def stream_logs(self):
        """Stream logs from the FinetuningRun."""
        TrainingLogsConsumer(self).stream_logs()

    @property
    def blueprint_url(self) -> str:
        """
        Link to view this FinetuningRun in the Blueprint UI

        **Example:**

        ```python
        from baseten.training import FinetuningRun

        my_run = FinetuningRun("RUN_ID")
        my_run.blueprint_url
        ```
        """
        return _fetch_blueprint_url(self.id, "trained-model")

    @property
    def blueprint_deployed_model_url(self) -> str:
        """
        Link to view the deployed model associated with this FinetuningRun in the Blueprint UI

        **Example:**

        ```python
        from baseten.training import FinetuningRun

        my_run = FinetuningRun("RUN_ID")
        my_run.blueprint_deployed_model_url
        ```
        """
        return _fetch_blueprint_url(self.id, "deployed-model")

    @property
    def deployed_model(self) -> Optional[FoundationalModel]:
        """
        The deployed model associated with this training run.
        """
        if self.deployed_model_id:
            return _fetch_model_pipeline(self.truss_name, self.deployed_model_id)
        return None


def _fetch_blueprint_url(id: str, resource_type: str) -> str:
    """
    Link to view this FinetuningRun in the Blueprint UI

    **Example:**

    ```python
    from baseten.training import FinetuningRun

    my_run = FinetuningRun("RUN_ID")
    my_run.blueprint_url
    ```
    """
    try:
        blueprint_project_id = get_blueprint_projects()[0]["id"]
        return (
            f"{settings.get_server_url()}/blueprint/projects/"
            f"{blueprint_project_id}?est={resource_type}--{id}"
        )
    except IndexError:
        raise ValueError("User does not have any blueprint projects.")


def _fetch_model_pipeline(truss_name: str, model_id: str) -> FoundationalModel:
    truss_name_to_inference_pipeline = {
        "dreambooth": lambda model_id: StableDiffusionPipeline(model_id=model_id),
        "full_stable_diffusion": lambda model_id: StableDiffusionPipeline(model_id=model_id),
        "flan_t5_base": lambda model_id: FlanT5(model_id=model_id),
        "llama7b": lambda model_id: Llama(model_id=model_id),
    }

    return truss_name_to_inference_pipeline[truss_name](model_id)



from typing import Optional

from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.models.foundational_model import FoundationalModel
from baseten.models.util import requests_error_handling


class Llama(FoundationalModel):
    """
    LLaMA is a text-to-text generation model from Meta.

    !!! note
        This `Llama` model object is only available as the interface for a fine-tuned model.

    **Examples:**

    ```python
    model = Llama(model_id="llama-1234")
    completion = model("What is the meaning of life?", max_length=256)
    ```
    """

    def __init__(self, model_id: Optional[str] = None):
        """
        Args:
            model_id: The ID for a deployed model created from a LLaMA fine-tuning run.
        """

        self._id = model_id
        self._is_user_model = self._id is not None
        self._model = self._set_user_model()

    def status(self) -> str:
        return self._model.status

    def id(self) -> str:
        return self._model.id

    def _set_user_model(self) -> BasetenDeployedModel:
        """Creates internal BasetenDeployedModel object that points to users
        deployed LLaMA model. If the user does not have a deployed model,
        we will create one for them.
        """
        if self._is_user_model:
            return BasetenDeployedModel(model_id=self._id)
        else:
            raise ValueError("No model id provided")

    def __call__(self, prompt, seed=None, **kwargs) -> str:
        """Generate text from a prompt. Supports all parameters of the underlying model
        from Huggingface library. The below are _some_ of the parameters that can be passed in.
        See Huggingface Transformers `.generate()` documentation for more details.

        Args:
            prompt (str): The prompt to generate text from.
            seed (int): The random seed to use for reproducibility. Optional.
        Attributes:
            num_beams (int): Number of beams for beam search. 1 means no beam search. Optional. Default to 1.
            num_return_sequences (int): The number of independently computed returned sequences for
                each element in the batch. Default to 1.
            max_length (int): The maximum number of tokens that can be generated.
            temperature (float): The value used to module the next token probabilities. Must be
                strictly positive. Default to 1.0.
            top_k (int): The number of highest probability vocabulary tokens to keep for top-k-filtering.
                Between 1 and infinity. Default to 50.
            top_p (float): If set to float < 1, only the most probable tokens with probabilities that
                add up to top_p or higher are kept for generation. Default to 1.0.
            repetition_penalty (float): The parameter for repetition penalty. Between 1.0 and infinity.
                1.0 means no penalty. Default to 1.0.
            length_penalty (float): Exponential penalty to the sequence length. Default to 1.0.
            no_repeat_ngram_size (int): If set to int > 0, all ngrams of that size can only occur
                once. Default to 0.
            num_return_sequences (int): The number of independently computed returned sequences
                for each element in the batch. Default to 1.
            do_sample (bool): If set to False greedy decoding is used. Otherwise sampling is used.
                Defaults to True.
            early_stopping (bool): Whether to stop the beam search when at least num_beams sentences
                are finished per batch or not. Default to False.
            use_cache (bool): Whether or not the model should use the past last key/values attentions
                (if applicable to the model) to speed up decoding. Default to True.
            decoder_start_token_id (int): If an encoder-decoder model starts decoding with a different
                token than BOS, the id of that token. Default to None.
            pad_token_id (int): The id of the padding token. Default to None.
            eos_token_id (int): The id of the end of sequence token. Default to None.
            forced_bos_token_id (int): The id of the token to force as the first generated
                token after the BOS token. Default to None.
            forced_eos_token_id (int): The id of the token to force as the last generated
                token when max_length is reached. Default to None.
            remove_invalid_values (bool): Whether or not to remove possible `nan` and `inf`
                outputs of the model to prevent the generation method to crash. Default to False.

        Returns:
            generated_text (str): The generated text.
        """
        request_body = {"prompt": prompt, **kwargs}

        with requests_error_handling():
            server_response = self._model.predict(request_body)

        if server_response["status"] == "error":
            raise ValueError(server_response["message"])
        return server_response["data"]

    @staticmethod
    def url():
        raise ValueError("Llama is only supported as a user-finetuned model")



from dataclasses import dataclass
from typing import Optional


@dataclass
class TryPatchDraftTrussResponse:
    succeeded: bool
    needs_full_deploy: bool
    model_version_info: Optional[dict] = None
    error_message: Optional[str] = None
    success_message: Optional[str] = None

    @staticmethod
    def needs_full_deploy_response(
        error_message: Optional[str] = None,
    ) -> "TryPatchDraftTrussResponse":
        return TryPatchDraftTrussResponse(
            succeeded=False,
            needs_full_deploy=True,
            error_message=error_message,
        )

    @staticmethod
    def success_response(model_version_info: dict, success_message: Optional[str] = None):
        return TryPatchDraftTrussResponse(
            succeeded=True,
            needs_full_deploy=False,
            model_version_info=model_version_info,
            success_message=success_message,
        )



import functools
import json
import logging
import os
from typing import IO, Dict, List, Optional, Tuple, Union

import boto3
import requests
from boto3.s3.transfer import TransferConfig
from colorama import Fore
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm

from baseten.common import settings
from baseten.common.core import ApiError, AuthorizationError
from baseten.common.util import base64_encoded_json_str

logger = logging.getLogger(__name__)


class AuthToken:
    value: str

    def __init__(self, value: str):
        self.value = value

    def headers(self):
        raise NotImplementedError


class ApiKey(AuthToken):
    def headers(self):
        return {"Authorization": f"Api-Key {self.value}"}


class JWT(AuthToken):
    def headers(self):
        return {"Authorization": self.value}


def with_api_key(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = settings.read_config()

        if config.has_option("api", "api_key"):
            api_key = config.get("api", "api_key")
        elif settings.get_jwt_auth_token():
            raise AuthorizationError("This operation is not supported in the Web IDE")
        else:
            raise AuthorizationError("You must first run the `baseten login` cli command.")

        result = func(ApiKey(api_key), *args, **kwargs)
        return result

    return wrapper


def with_api_key_or_jwt(func):
    """
    Allows an API method to be called internally (with JWT) or externally
    (with API Key).
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = settings.read_config()

        if config.has_option("api", "api_key"):
            result = func(ApiKey(config.get("api", "api_key")), *args, **kwargs)
            return result
        elif settings.get_jwt_auth_token():
            result = func(JWT(settings.get_jwt_auth_token()), *args, **kwargs)
            return result
        else:
            raise AuthorizationError("You must first run the `baseten login` cli command.")

    return wrapper


@with_api_key_or_jwt
def models(auth_token: AuthToken):
    query_string = """
    {
      models {
        id,
        name
        versions{
            id,
            semver,
            current_deployment_status,
            is_primary,
        }
      }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]


@with_api_key
def get_model(api_key, model_name):
    query_string = f"""
    {{
      model_version(name: "{model_name}") {{
        oracle{{
            name
            versions{{
                id
                semver
                truss_hash
                truss_signature
                is_draft
                current_model_deployment_status {{
                    status
                }}
            }}
        }}
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]


@with_api_key_or_jwt
def get_primary_model_version_status(api_key, model_id: str):
    query_string = f"""
    {{
      model(id: "{model_id}") {{
        primary_version{{
            current_model_deployment_status {{
                status
            }}
        }}
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model"]["primary_version"]["current_model_deployment_status"]["status"]


@with_api_key_or_jwt
def get_model_version_status(api_key, model_version_id: str):
    query_string = f"""
    {{
        model_version(id: "{model_version_id}") {{
            current_model_deployment_status {{
                status
            }}
        }}
    }}

    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_version"]["current_model_deployment_status"]["status"]


@with_api_key
def get_pretrained_model(api_key, pretrained_model_name):
    query_string = f"""
    {{
      pretrained_model(name: "{pretrained_model_name}") {{
        pretty_name
        name
        s3_key
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]


@with_api_key
def pretrained_models(api_key):
    query_string = """
    {
      pretrained_models {
        pretty_name
        name
        s3_key
      }
    }
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]


@with_api_key_or_jwt
def pretrained_model_version(auth_token: AuthToken, pretrained_model_name: str):
    query_string = f"""
    {{
      pretrained_model_version(pretrained_model_name: "{pretrained_model_name}"){{
        model_id
        current_deployment_status
        truss_spec_version
      }}
    }}
    """
    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]


@with_api_key
def models_summary(api_key):
    query_string = """
    {
      models {
        id,
        name,
        description,
        versions {
          id
          created
        }
      }
    }
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["models"]


@with_api_key
def model_version_external_id_get_version(api_key, external_model_version_id):
    query_string = f"""
    {{
      model_version(external_model_version_id: "{external_model_version_id}") {{
        id
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_version"]["id"]


@with_api_key_or_jwt
def model_version_truss_spec_version(auth_token: AuthToken, model_version_id: str):
    query_string = f"""
    {{
      model_version(id: "{model_version_id}") {{
        truss_spec_version
      }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["model_version"]["truss_spec_version"]


@with_api_key_or_jwt
def model_truss_spec_version(auth_token: AuthToken, model_id: str):
    query_string = f"""
    {{
      model(id: "{model_id}") {{
        primary_version {{
            truss_spec_version
        }}
      }}
    }}
    """
    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["model"]["primary_version"]["truss_spec_version"]


@with_api_key_or_jwt
def model_id_from_model_version(auth_token: AuthToken, model_version_id: str) -> str:
    query_string = f"""
    {{
      model_version(id: "{model_version_id}") {{
        oracle {{
            id
        }}
      }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["model_version"]["oracle"]["id"]


@with_api_key
def create_model_from_truss(
    api_key,
    model_name,
    s3_key,
    config,
    semver_bump,
    client_version,
    is_trusted=False,
    external_model_version_id=None,
):
    query_string = f"""
    mutation {{
      create_model_from_truss(name: "{model_name}",
                   s3_key: "{s3_key}",
                   config: "{config}",
                   semver_bump: "{semver_bump}",
                   client_version: "{client_version}",
                   is_trusted: {'true' if is_trusted else 'false'}
                   external_model_version_id: "{external_model_version_id if external_model_version_id else ''}"
) {{
        id,
        name,
        version_id
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_model_from_truss"]


@with_api_key
def create_model_version_from_truss(
    api_key,
    model_id,
    s3_key,
    config,
    semver_bump,
    client_version,
    is_trusted=False,
    external_model_version_id=None,
):
    query_string = f"""
    mutation {{
      create_model_version_from_truss(
                   model_id: "{model_id}"
                   s3_key: "{s3_key}",
                   config: "{config}",
                   semver_bump: "{semver_bump}",
                   client_version: "{client_version}",
                   is_trusted: {'true' if is_trusted else 'false'}
                   external_model_version_id: "{external_model_version_id if external_model_version_id else ''}"

      ) {{
        id,
        name,
        version_id
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_model_version_from_truss"]


@with_api_key
def deploy_draft_truss(
    api_key,
    model_name,
    s3_key,
    config,
    client_version,
    is_trusted=False,
):
    query_string = f"""
    mutation {{
      deploy_draft_truss(name: "{model_name}",
                   s3_key: "{s3_key}",
                   config: "{config}",
                   client_version: "{client_version}",
                   is_trusted: {'true' if is_trusted else 'false'},
) {{
        id,
        name,
        version_id
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["deploy_draft_truss"]


@with_api_key
def patch_draft_truss(
    api_key,
    model_name,
    patch,
    client_version,
):
    patch = base64_encoded_json_str(patch.to_dict())
    query_string = f"""
    mutation {{
      patch_draft_truss(name: "{model_name}",
                   client_version: "{client_version}",
                   patch: "{patch}",
) {{
        id,
        name,
        version_id
        succeeded
        needs_full_deploy
        error
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["patch_draft_truss"]


@with_api_key
def signed_s3_upload_post(api_key, model_file_name):
    query_string = f"""
    {{
      signed_s3_upload_url(model_file_name: "{model_file_name}") {{
        url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["signed_s3_upload_url"]


@with_api_key
def model_s3_upload_credentials(api_key):
    query_string = """
    {
        model_s3_upload_credentials {
            s3_bucket
            s3_key
            aws_access_key_id
            aws_secret_access_key
            aws_session_token
        }
    }
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_s3_upload_credentials"]


@with_api_key
def signed_s3_upload_and_download_post(api_key, file_name):
    query_string = f"""
    {{
      signed_s3_upload_url(model_file_name: "{file_name}"){{
        url,
        get_url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["signed_s3_upload_url"]


@with_api_key
def get_model_s3_download_url(api_key, model_version_id):
    query_string = f"""
    {{
      model_s3_download_url(model_version_id: "{model_version_id}") {{
        url,
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["model_s3_download_url"]["url"]


@with_api_key_or_jwt
def user_files_signed_s3_upload_post(auth_token: AuthToken, file_name: str):
    query_string = f"""
    {{
      user_files_signed_s3_upload_url(file_name: "{file_name}") {{
        url,
        get_url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)

    return resp["data"]["user_files_signed_s3_upload_url"]


@with_api_key
def register_data_for_model(api_key, s3_key, model_version_id, data_name):
    query_string = f"""
    mutation {{
        create_sample_data_file(model_version_id: "{model_version_id}",
                                name: "{data_name}",
                                s3_key: "{s3_key}") {{
          id
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_sample_data_file"]


@with_api_key_or_jwt
def predict_for_model(
    auth_token: AuthToken,
    model_id: str,
    inputs: Union[List, Dict],
    metadata: Optional[List[Dict]] = None,
    prediction_only: bool = True,
    v2: bool = False,
) -> Union[List[List], Dict]:
    """Call the model's predict given the input json.

    Args:
        auth_token (AuthToken)
        model_id (str)
        inputs (List)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.
        prediction_only(bool)

    Raises:
        RequestException: If there was an error communicating with the server.
    """

    # We have a special predict endpoint for internal use that we use
    # if a JWT is present.
    if isinstance(auth_token, ApiKey):
        predict_url = f"{settings.get_server_url()}/models/{model_id}/predict"
    elif isinstance(auth_token, JWT):
        predict_url = f"{settings.get_server_url()}/models/{model_id}/predict_internal"

    return _predict(auth_token, predict_url, inputs, metadata, prediction_only, v2=v2)


@with_api_key
def predict_for_model_version(
    api_key: ApiKey,
    model_version_id: str,
    inputs: Union[List, Dict],
    metadata: Optional[List[Dict]] = None,
    prediction_only: bool = True,
    v2: bool = False,
) -> Union[List[List], Dict]:
    """Call the model version's predict given the input json.

    Args:
        api_key (str)
        model_version_id (str)
        inputs (List)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    predict_url = f"{settings.get_server_url()}/model_versions/{model_version_id}/predict"
    return _predict(api_key, predict_url, inputs, metadata, prediction_only, v2=v2)


@with_api_key
def set_primary(api_key, model_version_id: str):
    """Promote this version of the model as the primary version.

    Args:
        api_key (str)
        model_version_id (str)
    """
    query_string = f"""
    mutation {{
      update_model_version(model_version_id: "{model_version_id}", is_primary: true) {{
        id,
        is_primary,
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["update_model_version"]


@with_api_key
def update_model_features(
    api_key: ApiKey, model_version_id: str, feature_names: List, class_labels: Optional[List] = None
):
    """Update the feature names for the model.

    Args:
        api_key (str)
        model_version_id (str)
        feature_names (List)
        class_labels (Optional[List]): applies only to classifiers.
    """
    encoded_feature_names = base64_encoded_json_str(feature_names)
    encoded_class_labels = base64_encoded_json_str(class_labels)
    query_string = f"""
    mutation {{
      update_model_version(model_version_id: "{model_version_id}",
                           encoded_feature_names: "{encoded_feature_names}",
                           encoded_class_labels: "{encoded_class_labels}") {{
        id,
        feature_names,
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["update_model_version"]


@with_api_key
def install_requirements(api_key, requirements_txt):
    escaped_requirements_txt = requirements_txt.replace(
        "\n", "\\n"
    )  # Otherwise the mutation becomes invalid graphql.
    query_string = f"""
    mutation {{
      create_pynode_requirement(requirements_txt: "{escaped_requirements_txt}") {{
        id
        status
        error_message
      }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["create_pynode_requirement"]


@with_api_key
def requirement_status(api_key, requirement_id):
    query_string = f"""
    {{
      pynode_requirement(id: "{requirement_id}") {{
        id
        status
        error_message
      }}
    }}
    """

    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["pynode_requirement"]


def _format_inputs_for_v1(
    auth_token: AuthToken, inputs: Union[List, Dict], metadata: Optional[List[Dict]]
) -> Union[List, Dict]:
    """
    For internal (JWT) token use-cases, we do not need to package up the inputs.
    For external cases, for the v1 model serving path, we expect inputs to be
    provided in the format { inputs: ..., metadata: ... }.
    """
    if isinstance(auth_token, ApiKey):
        return {"inputs": inputs, "metadata": metadata}
    else:
        return inputs


def _predict(
    auth_token: AuthToken,
    predict_url: str,
    inputs: Union[List, Dict],
    metadata: Optional[List[Dict]] = None,
    prediction_only: bool = True,
    v2: bool = False,
):
    if v2:
        resp = _post_rest_query(auth_token, predict_url, inputs)
        resp_json = json.loads(resp.content)
        return resp_json["model_output"]

    resp = _post_rest_query(
        auth_token, predict_url, _format_inputs_for_v1(auth_token, inputs, metadata)
    )
    resp_json = json.loads(resp.content)
    return resp_json["predictions"] if prediction_only else resp_json


def _post_graphql_query(auth_token: AuthToken, query_string: str) -> dict:
    headers = auth_token.headers()

    resp = requests.post(
        f"{settings.get_server_url()}/graphql/", data={"query": query_string}, headers=headers
    )
    if not resp.ok:
        logger.error(f"GraphQL endpoint failed with error: {resp.content}")  # type: ignore
        resp.raise_for_status()
    resp_dict = resp.json()
    errors = resp_dict.get("errors")
    if errors:
        raise ApiError(errors[0]["message"], resp)
    return resp_dict


def _post_rest_query(auth_token: AuthToken, url: str, post_body_dict: Union[List, Dict]):
    headers = auth_token.headers()

    resp = requests.post(url, json=post_body_dict, headers=headers)
    resp.raise_for_status()
    return resp


def upload_model(serialize_file: IO, file_ext: str, file_name: str) -> str:
    """Uploads the serialized model to the appropriate environment

    Args:
        serialize_file (file): A file-like object that is the serialized representation of the model object.
        file_ext (str): The file extension for the saved model

    Returns:
        str: The key for the uploaded model

    Raises:
        RequestException: If there was an error communicating with the server.
    """

    temp_credentials_s3_upload = model_s3_upload_credentials()
    s3_key = temp_credentials_s3_upload.pop("s3_key")
    s3_bucket = temp_credentials_s3_upload.pop("s3_bucket")
    logger.info("🚀 Uploading model to Baseten 🚀")
    _multipart_upload_boto3(serialize_file.name, s3_bucket, s3_key, temp_credentials_s3_upload)
    return s3_key


def upload_user_file(io_stream: IO, file_name: str) -> str:
    signed_s3_upload_post = user_files_signed_s3_upload_post(file_name)
    logger.debug(f"Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}")

    form_fields = signed_s3_upload_post["form_fields"]
    form_fields["AWSAccessKeyId"] = form_fields.pop(
        "aws_access_key_id"
    )  # S3 expects key name AWSAccessKeyId
    form_fields["file"] = (file_name, io_stream)

    _upload_file(signed_s3_upload_post, form_fields)

    return signed_s3_upload_post["get_url"]


def upload_user_dataset_file(file_path: str, file_name: str) -> dict:
    signed_s3_upload_post = user_files_signed_s3_upload_post(file_name)
    logger.debug(f"Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}")

    form_fields = signed_s3_upload_post["form_fields"]
    form_fields["AWSAccessKeyId"] = form_fields.pop(
        "aws_access_key_id"
    )  # S3 expects key name AWSAccessKeyId
    form_fields["file"] = (file_name, open(file_path, "rb"))

    _upload_file(signed_s3_upload_post, form_fields)

    return signed_s3_upload_post


def _multipart_upload_boto3(file_path, bucket_name, key, credentials):
    s3_resource = boto3.resource("s3", **credentials)
    filesize = os.stat(file_path).st_size

    with tqdm(
        total=filesize,
        desc="Upload",
        unit="B",
        unit_scale=True,
    ) as pbar:
        s3_resource.Object(bucket_name, key).upload_file(
            file_path,
            Config=TransferConfig(
                max_concurrency=10,
                use_threads=True,
            ),
            Callback=pbar.update,
        )


def _upload_file(this_signed_s3_upload_post: dict, form_fields: dict) -> str:
    encoder = MultipartEncoder(fields=form_fields)
    encoder_len = encoder.len
    pbar = tqdm(
        total=encoder_len,
        unit_scale=True,
        bar_format="Upload Progress: {percentage:3.0f}%% |%s{bar:100}%s| {n_fmt}/{total_fmt}"
        % (Fore.BLUE, Fore.RESET),
    )

    def callback(monitor):
        progress = monitor.bytes_read - pbar.n
        pbar.update(progress)

    monitor = MultipartEncoderMonitor(encoder, callback)
    resp = requests.post(
        this_signed_s3_upload_post["url"],
        data=monitor,
        headers={"Content-Type": monitor.content_type},
    )
    resp.raise_for_status()
    pbar.close()
    logger.info("🔮 Upload successful!🔮")

    logger.debug(f"File upload HTTP status code: {resp.status_code} and content:\n{resp.content}")  # type: ignore

    return this_signed_s3_upload_post["form_fields"]["key"]


@with_api_key
def deactivate_model_version(api_key, model_version_id):
    query_string = f"""
    mutation {{
        deactivate_model_version(model_version_id: "{model_version_id}") {{
          ok
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["deactivate_model_version"]["ok"]


@with_api_key
def activate_model_version(api_key, model_version_id):
    query_string = f"""
    mutation {{
        activate_model_version(model_version_id: "{model_version_id}") {{
          ok
        }}
    }}
    """
    resp = _post_graphql_query(api_key, query_string)
    return resp["data"]["activate_model_version"]["ok"]


@with_api_key_or_jwt
def create_pretrained_model(auth_token: AuthToken, model_zoo_name: str, model_name: str):
    query_string = f"""
    mutation {{
        create_pretrained_model(
            model_zoo_name: "{model_zoo_name}",
            name: "{model_name}",
            description : "",
            create_app: false
        ) {{
            model {{
                created
                id
                name
                platformType: model_platform_type
                numberOfVersions: number_of_versions
                description: description
                primaryVersion: primary_version {{
                    id
                    model_id
                    modelFramework: model_framework_display_name
                    truss_spec_version
                }}
            }}
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["create_pretrained_model"]["model"]


def _generate_graphql_params(params: dict) -> str:
    """
    Generates a GraphQL parameter string from a dictionary
    """
    return ",".join([f'{key}: "{value}"' for key, value in params.items()])


@with_api_key_or_jwt
def create_training_dataset(
    auth_token: AuthToken, s3_key: str, training_type: str, dataset_name: Optional[str]
) -> str:
    # Only include the dataset name if it is passed.
    graphql_params = {"training_type": training_type, "s3_key": s3_key}

    if dataset_name is not None:
        graphql_params.update({"name": dataset_name})

    query_string = f"""
    mutation {{
        create_training_dataset(
            {_generate_graphql_params(graphql_params)}
        ) {{
            id
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["create_training_dataset"]["id"]


@with_api_key_or_jwt
def get_training_dataset(auth_token: AuthToken, dataset_id: str) -> dict:
    query_string = f"""
    {{
        training_dataset(id: "{dataset_id}") {{
            name
            id
            presigned_s3_get_url
            training_type
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["training_dataset"]


@with_api_key_or_jwt
def get_training_run_by_id(auth_token: AuthToken, training_run_id: str) -> Dict:
    query_string = f"""
    {{
        training_run(id: "{training_run_id}") {{
            id
            status
            trained_model_name

            variables {{
                key
                value
            }}
            created
            started
            stopped
            truss_name
            latest_oracle_version {{
                id
                oracle {{
                    id
                }}
            }}
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["training_run"]


@with_api_key_or_jwt
def get_training_run_by_name(auth_token: AuthToken, trained_model_name: str) -> Dict:
    query_string = f"""
    {{
        training_run_by_trained_model_name(trained_model_name: "{trained_model_name}") {{
            id
            status
            trained_model_name
            variables {{
                key
                value
            }}
            created
            started
            stopped
            truss_name
            latest_oracle_version {{
                id
                oracle {{
                    id
                }}
            }}
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["training_run_by_trained_model_name"]


@with_api_key_or_jwt
def get_all_training_runs(auth_token: AuthToken) -> List[Dict]:
    query_string = """
    query {
        training_runs {
            id
            trained_model_name
        }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["training_runs"]


@with_api_key_or_jwt
def finetune_zoo_model(
    auth_token: AuthToken,
    trained_model_name: str,
    train_truss_name: str,
    encoded_variables: str,
    auto_deploy: bool,
) -> str:
    query_string = f"""
    mutation {{
        finetune_model_zoo_model(
            truss_name: "{train_truss_name}",
            encoded_variables: "{encoded_variables}",
            trained_model_name: "{trained_model_name}",
            auto_deploy: {json.dumps(auto_deploy)}
        ) {{
            id
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["finetune_model_zoo_model"]["id"]


@with_api_key_or_jwt
def deploy_from_training_run(
    auth_token: AuthToken,
    training_run_id: str,
    name: str,
    idle_time_minutes: int,
) -> Tuple[str, str]:
    query_string = f"""
    mutation {{
        deploy_from_training_run(
            training_run_id: "{training_run_id}",
            name: "{name}",
            idle_time_minutes: {idle_time_minutes},
        ) {{
            version_id
            id

        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    deploy_data = resp["data"]["deploy_from_training_run"]
    return deploy_data["id"], deploy_data["version_id"]


@with_api_key_or_jwt
def cancel_training_run(
    auth_token: AuthToken,
    training_run_id: str,
) -> bool:
    query_string = f"""
    mutation {{
        cancel_training_run(
            training_run_id: "{training_run_id}",
        ) {{
            cancelled
        }}
    }}
    """

    resp = _post_graphql_query(auth_token, query_string)
    deploy_data = resp["data"]["cancel_training_run"]
    return deploy_data["cancelled"]


@with_api_key_or_jwt
def get_training_logs(
    auth_token: AuthToken,
    training_run_id: str,
    start: Optional[str] = None,
    end: Optional[str] = None,
    limit: Optional[int] = None,
    direction: Optional[str] = None,
) -> List[dict]:
    query_string = f"""
    query {{
        logs(
            log_type: TRAINING,
            entity_id: "{training_run_id}",
            {f'start: "{start}",' if start is not None else ""}
            {f'end: "{end}",' if end is not None else ""}
            {f'limit: {limit},' if limit is not None else ""}
            {f'direction: "{direction}",' if direction is not None else ""}
        ) {{
            ts
            level
            msg
        }}
    }}
    """
    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["logs"]


@with_api_key_or_jwt
def get_blueprint_projects(
    auth_token: AuthToken,
) -> List[dict]:
    query_string = """
    query {
        blueprint_workflows {
            id
            name
        }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["blueprint_workflows"]


@with_api_key_or_jwt
def get_organization_credits(auth_token: AuthToken) -> List[dict]:
    query_string = """
    query {
        organization {
            monetary_credit_granted
            monetary_credit_balance
            payment_method_status
        }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["organization"]


@with_api_key_or_jwt
def get_organization_scalable_to_zero(auth_token: AuthToken) -> bool:
    query_string = """
    query {
        organization {
            scale_new_models_to_zero
        }
    }
    """

    resp = _post_graphql_query(auth_token, query_string)
    return resp["data"]["organization"]["scale_new_models_to_zero"]



from baseten.models import flan_t5, llama, stable_diffusion, whisper

StableDiffusionPipeline = stable_diffusion.StableDiffusionPipeline
FlanT5 = flan_t5.FlanT5
Whisper = whisper.Whisper
Llama = llama.Llama
